# Copyright (c) 2016-2022 Memgraph Ltd. [https://memgraph.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import List, Optional, Union

from gqlalchemy.connection import Connection, Neo4jConnection
from gqlalchemy.exceptions import (
    GQLAlchemyError,
    GQLAlchemyUniquenessConstraintError,
)
from gqlalchemy.models import (
    Neo4jConstraintExists,
    Neo4jConstraintUnique,
    Neo4jIndex,
    Node,
    Relationship,
)
from gqlalchemy.vendors.database_client import DatabaseClient

__all__ = ("Neo4j",)

NEO4J_HOST = os.getenv("NEO4J_HOST", "localhost")
NEO4J_PORT = int(os.getenv("NEO4J_PORT", "7687"))
NEO4J_USERNAME = os.getenv("NEO4J_USERNAME", "neo4j")
NEO4J_PASSWORD = os.getenv("NEO4J_PASSWORD", "test")
NEO4J_ENCRYPTED = os.getenv("NEO4J_ENCRYPT", "false").lower() == "true"
NEO4J_CLIENT_NAME = os.getenv("NEO4J_CLIENT_NAME", "neo4j")


class Neo4jConstants:
    CONSTRAINT_TYPE = "constraint type"
    EXISTS = "exists"
    LABEL = "labelsOrTypes"
    PROPERTY = "property"
    PROPERTIES = "properties"
    UNIQUE = "unique"
    LOOKUP = "LOOKUP"
    TYPE = "type"
    UNIQUE = "UNIQUE"
    UNIQUENESS = "uniqueness"


class Neo4j(DatabaseClient):
    def __init__(
        self,
        host: str = NEO4J_HOST,
        port: int = NEO4J_PORT,
        username: str = NEO4J_USERNAME,
        password: str = NEO4J_PASSWORD,
        encrypted: bool = NEO4J_ENCRYPTED,
        client_name: str = NEO4J_CLIENT_NAME,
    ):
        super().__init__(
            host=host, port=port, username=username, password=password, encrypted=encrypted, client_name=client_name
        )
        self._cached_connection: Optional[Connection] = None

    def get_indexes(self) -> List[Neo4jIndex]:
        """Returns a list of all database indexes (label and label-property types)."""
        indexes = []
        for result in self.execute_and_fetch("SHOW INDEX;"):
            indexes.append(
                Neo4jIndex(
                    result[Neo4jConstants.LABEL][0]
                    if result[Neo4jConstants.TYPE] != Neo4jConstants.LOOKUP
                    else result[Neo4jConstants.LABEL],
                    result[Neo4jConstants.PROPERTIES][0]
                    if result[Neo4jConstants.TYPE] != Neo4jConstants.LOOKUP
                    else result[Neo4jConstants.PROPERTIES],
                    result[Neo4jConstants.TYPE],
                    result[Neo4jConstants.UNIQUENESS],
                )
            )
        return indexes

    def ensure_indexes(self, indexes: List[Neo4jIndex]) -> None:
        """Ensures that database indexes match input indexes."""
        old_indexes = set(self.get_indexes())
        new_indexes = set(indexes)
        for obsolete_index in old_indexes.difference(new_indexes):
            if obsolete_index.type != Neo4jConstants.LOOKUP and obsolete_index.uniqueness != Neo4jConstants.UNIQUE:
                self.drop_index(obsolete_index)
        for missing_index in new_indexes.difference(old_indexes):
            self.create_index(missing_index)

    def get_constraints(
        self,
    ) -> List[Union[Neo4jConstraintExists, Neo4jConstraintUnique]]:
        """Returns a list of all database constraints (label and label-property types)."""
        constraints: List[Union[Neo4jConstraintExists, Neo4jConstraintUnique]] = []
        for result in self.execute_and_fetch("SHOW CONSTRAINTS;"):
            if result[Neo4jConstants.TYPE] == "UNIQUENESS":
                constraints.append(
                    Neo4jConstraintUnique(
                        result[Neo4jConstants.LABEL][0],
                        tuple(result[Neo4jConstants.PROPERTIES]),
                    )
                )
        return constraints

    def get_exists_constraints(
        self,
    ) -> List[Neo4jConstraintExists]:
        return [x for x in self.get_constraints() if isinstance(x, Neo4jConstraintExists)]

    def get_unique_constraints(
        self,
    ) -> List[Neo4jConstraintUnique]:
        return [x for x in self.get_constraints() if isinstance(x, Neo4jConstraintUnique)]

    def new_connection(self) -> Connection:
        """Creates new Neo4j connection."""
        args = dict(
            host=self._host,
            port=self._port,
            username=self._username,
            password=self._password,
            encrypted=self._encrypted,
            client_name=self._client_name,
        )
        return Neo4jConnection(**args)

    def save_node(self, node: Node) -> Node:
        """Saves node to the database.
        If the node._id is not None it fetches the node with the same id from
        the database and updates it's fields.
        If the node has unique fields it fetches the nodes with the same unique
        fields from the database and updates it's fields.
        Otherwise it creates a new node with the same properties.
        Null properties are ignored.
        """
        result = None
        if node._id is not None:
            result = self.save_node_with_id(node)
        elif node.has_unique_fields():
            matching_nodes = list(self._get_nodes_with_unique_fields(node))
            if len(matching_nodes) > 1:
                raise GQLAlchemyUniquenessConstraintError(
                    f"Uniqueness constraints match multiple nodes: {matching_nodes}"
                )
            elif len(matching_nodes) == 1:
                node._id = matching_nodes[0]["node"]._id
                result = self.save_node_with_id(node)
            else:
                result = self.create_node(node)
        else:
            result = self.create_node(node)

        return result

    def load_node(self, node: Node) -> Optional[Node]:
        """Loads a node from the database.
        If the node._id is not None it fetches the node from the database with that
        internal id.
        If the node has unique fields it fetches the node from the database with
        those unique fields set.
        Otherwise it tries to find any node in the database that has all properties
        set to exactly the same values.
        If no node is found or no properties are set it raises a GQLAlchemyError.
        """
        if node._id is not None:
            result = self.load_node_with_id(node)
        elif node.has_unique_fields():
            matching_node = self.get_variable_assume_one(
                query_result=self._get_nodes_with_unique_fields(node), variable_name="node"
            )
            result = matching_node
        else:
            result = self.load_node_with_all_properties(node)

        return result

    def load_relationship(self, relationship: Relationship) -> Optional[Relationship]:
        """Returns a relationship loaded from the database.
        If the relationship._id is not None it fetches the relationship from
        the database that has the same internal id.
        Otherwise it returns the relationship whose relationship._start_node_id
        and relationship._end_node_id and all relationship properties that
        are not None match the relationship in the database.
        If there is no relationship like that in the database, or if there are
        multiple relationships like that in the database, throws GQLAlchemyError.
        """
        if relationship._id is not None:
            result = self.load_relationship_with_id(relationship)
        elif relationship._start_node_id is not None and relationship._end_node_id is not None:
            result = self.load_relationship_with_start_node_id_and_end_node_id(relationship)
        else:
            raise GQLAlchemyError("Can't load a relationship without a start_node_id and end_node_id.")
        return result

    def save_relationship(self, relationship: Relationship) -> Optional[Relationship]:
        """Saves a relationship to the database.
        If relationship._id is not None it finds the relationship in the database
        and updates it's properties with the values in `relationship`.
        If relationship._id is None, it creates a new relationship.
        If you want to set a relationship._id instead of creating a new
        relationship, use `load_relationship` first.
        """
        if relationship._id is not None:
            result = self.save_relationship_with_id(relationship)
        elif relationship._start_node_id is not None and relationship._end_node_id is not None:
            result = self.create_relationship(relationship)
        else:
            raise GQLAlchemyError("Can't create a relationship without start_node_id and end_node_id.")

        return result



# Copyright (c) 2016-2022 Memgraph Ltd. [https://memgraph.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Union, Tuple, List
from string import ascii_lowercase

from gqlalchemy.query_builders.declarative_base import (  # noqa F401
    Call,
    Create,
    DeclarativeBase,
    DeclarativeBaseTypes,
    Foreach,
    Match,
    Merge,
    Operator,
    Order,
    PartialQuery,
    Return,
    Unwind,
    With,
)
from gqlalchemy.vendors.database_client import DatabaseClient
from gqlalchemy.vendors.memgraph import Memgraph
from gqlalchemy.utilities import CypherVariable, CypherNode, CypherRelationship, RelationshipDirection


class MemgraphQueryBuilderTypes(DeclarativeBaseTypes):
    LOAD_CSV = "LOAD_CSV"


class LoadCsvPartialQuery(PartialQuery):
    def __init__(self, path: str, header: bool, row: str):
        super().__init__(DeclarativeBaseTypes.LOAD_CSV)
        self.path = path
        self.header = header
        self.row = row

    def construct_query(self) -> str:
        return f" LOAD CSV FROM '{self.path}' " + ("WITH" if self.header else "NO") + f" HEADER AS {self.row} "


class QueryBuilder(DeclarativeBase):
    def __init__(self, connection: Optional[Memgraph] = None):
        super().__init__(connection)

    def load_csv(self, path: str, header: bool, row: str) -> "DeclarativeBase":
        """Load data from a CSV file by executing a Cypher query for each row.

        Args:
            path: A string representing the path to the CSV file. If beginning with `http://`, `https://`, or `ftp://`, the CSV file will be fetched over the network.
            header: A bool indicating if the CSV file starts with a header row.
            row: A string representing the name of the variable for iterating
              over each row.

        Returns:
            A `DeclarativeBase` instance for constructing queries.

        Examples:
            Load CSV with header:

            Python: `load_csv(path="path/to/my/file.csv", header=True, row="row").return_().execute()`
            Cypher: `LOAD CSV FROM 'path/to/my/file.csv' WITH HEADER AS row RETURN *;`

            Load CSV without header:

            Python: `load_csv(path='path/to/my/file.csv', header=False, row='row').return_().execute()`
            Cypher: `LOAD CSV FROM 'path/to/my/file.csv' NO HEADER AS row RETURN *;`
        """
        self._query.append(LoadCsvPartialQuery(path, header, row))

        return self

    def _construct_subgraph_path(
        self,
        relationship_types: Optional[Union[str, List[List[str]]]] = None,
        relationship_directions: Optional[
            Union[RelationshipDirection, List[RelationshipDirection]]
        ] = RelationshipDirection.RIGHT,
    ) -> str:
        """Constructs a MATCH query defining a subgraph using node labels and relationship types.

        Args:
            relationship_types: A string or list of lists of types of relationships used in the subgraph.
            relationship_directions: Enums representing directions.

        Returns:
            a string representing a MATCH query for a path with given node labels and relationship types.
        """

        query = f"{CypherNode(variable=ascii_lowercase[0])}"
        for i in range(len(relationship_types)):
            query += f"{CypherRelationship(relationship_types[i], relationship_directions[i])}{CypherNode()}"

        return query

    def call(
        self,
        procedure: str,
        arguments: Optional[Union[str, Tuple[Union[str, int, float]]]] = None,
        node_labels: Optional[Union[str, List[List[str]]]] = None,
        relationship_types: Optional[Union[str, List[List[str]]]] = None,
        relationship_directions: Optional[
            Union[RelationshipDirection, List[RelationshipDirection]]
        ] = RelationshipDirection.RIGHT,
        subgraph_path: str = None,
    ) -> "DeclarativeBase":
        """Override of base class method to support Memgraph's subgraph functionality.

        Method can be called with node labels and relationship types, both being optional, which are used to construct
        a subgraph, or if neither is provided, a subgraph query is used, which can be passed as a string representing a
        Cypher query defining the MATCH clause which selects the nodes and relationships to use.

        Args:
            procedure: A string representing the name of the procedure in the
              format `query_module.procedure`.
            arguments: A string representing the arguments of the procedure in
              text format.
            node_labels: Either a string, which is then used as the label for all nodes, or
                         a list of lists defining all labels for every node
            relationship_types: Types of relationships to be used in the subgraph. Either a
                                single type or a list of lists defining all types for every relationship
            relationship_directions: Directions of the relationships.
            subgraph_path: Optional way to define the subgraph via a Cypher MATCH clause.

        Returns:
            A `DeclarativeBase` instance for constructing queries.

        Examples:
            Python: `call('export_util.json', '/home/user', "LABEL", ["TYPE1", "TYPE2"]).execute()
            Cypher: `MATCH p=(a)-[:TYPE1 | :TYPE2]->(b) WHERE (a:LABEL) AND (b:LABEL)
                     WITH project(p) AS graph CALL export_util.json(graph, '/home/user')`

            or

            Python: `call('export_util.json', '/home/user', subgraph_path="(:LABEL)-[:TYPE]->(:LABEL)").execute()
            Cypher: `MATCH p=(:LABEL)-[:TYPE1]->(:LABEL) WITH project(p) AS graph
                    CALL export_util.json(graph, '/home/user')`
        """

        if not (node_labels is None and relationship_types is None):
            if isinstance(relationship_types, str):
                relationship_types = [[relationship_types]] * (
                    len(node_labels) - 1 if isinstance(node_labels, list) else 1
                )

            if isinstance(node_labels, str):
                node_labels = [[node_labels]] * (len(relationship_types) + 1 if relationship_types else 2)

            if isinstance(relationship_directions, RelationshipDirection):
                relationship_directions = [relationship_directions] * (
                    len(relationship_types) if relationship_types else 1
                )

            if (
                node_labels
                and relationship_types
                and (
                    len(node_labels) != len(relationship_types) + 1
                    or len(relationship_types) != len(relationship_directions)
                )
            ):
                raise ValueError(
                    "number of items in node_labels should be one more than in relationship_types and relationship_directions"
                )

            subgraph_path = f"{CypherNode(variable=ascii_lowercase[0])}"
            for i in range(len(relationship_directions)):
                rel_types = relationship_types[i] if relationship_types else None
                subgraph_path += f"{CypherRelationship(rel_types, relationship_directions[i])}{CypherNode(variable=ascii_lowercase[i + 1])}"

        if subgraph_path is not None:
            self._query.append(ProjectPartialQuery(subgraph_path=subgraph_path, node_labels=node_labels))

            if isinstance(arguments, str):
                arguments = (CypherVariable(name="graph"), arguments)
            elif isinstance(arguments, tuple):
                arguments = (CypherVariable(name="graph"), *arguments)
            else:
                arguments = CypherVariable(name="graph")

        super().call(procedure=procedure, arguments=arguments)

        return self


class LoadCsv(DeclarativeBase):
    def __init__(self, path: str, header: bool, row: str, connection: Optional[DatabaseClient] = None):
        super().__init__(connection)
        self._query.append(LoadCsvPartialQuery(path, header, row))


class ProjectPartialQuery(PartialQuery):
    def __init__(self, subgraph_path: str, node_labels: Optional[List[List[str]]] = None):
        super().__init__(DeclarativeBaseTypes.MATCH)
        self._subgraph_path = subgraph_path
        self._node_labels = node_labels

    @property
    def subgraph_path(self) -> str:
        return self._subgraph_path

    @property
    def node_labels(self) -> Optional[List[List[str]]]:
        return self._node_labels

    def construct_query(self) -> str:
        """Constructs a Project partial query.

        Given path part of a query (e.g. (:LABEL)-[:TYPE]->(:LABEL2)),
        adds MATCH, a path identifier and appends the WITH clause."""
        query = f" MATCH p={self.subgraph_path}"
        if self.node_labels:
            query += f" WHERE ({ascii_lowercase[0]}:" + f" or {ascii_lowercase[0]}:".join(self.node_labels[0]) + ")"
            for i in range(1, len(self.node_labels)):
                query += f" AND ({ascii_lowercase[i]}:" + f" or {ascii_lowercase[i]}:".join(self.node_labels[i]) + ")"

        return query + " WITH project(p) AS graph "



# Copyright (c) 2016-2022 Memgraph Ltd. [https://memgraph.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Any, Dict, Iterator, Optional

import mgclient
from neo4j import GraphDatabase
from neo4j.graph import Node as Neo4jNode
from neo4j.graph import Path as Neo4jPath
from neo4j.graph import Relationship as Neo4jRelationship

from gqlalchemy.exceptions import database_error_handler, connection_handler
from gqlalchemy.models import Node, Path, Relationship

__all__ = ("Connection",)


class Connection(ABC):
    def __init__(
        self,
        host: str,
        port: int,
        username: str,
        password: str,
        encrypted: bool,
        client_name: Optional[str] = None,
    ):
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.encrypted = encrypted
        self.client_name = client_name

    @abstractmethod
    def execute(self, query: str, parameters: Dict[str, Any] = {}) -> None:
        """Executes Cypher query without returning any results."""
        pass

    @abstractmethod
    def execute_and_fetch(self, query: str, parameters: Dict[str, Any] = {}) -> Iterator[Dict[str, Any]]:
        """Executes Cypher query and returns iterator of results."""
        pass

    @abstractmethod
    def is_active(self) -> bool:
        """Returns True if connection is active and can be used."""
        pass


class MemgraphConnection(Connection):
    def __init__(
        self,
        host: str,
        port: int,
        username: str,
        password: str,
        encrypted: bool,
        client_name: Optional[str] = None,
        lazy: bool = False,
    ):
        super().__init__(
            host=host, port=port, username=username, password=password, encrypted=encrypted, client_name=client_name
        )
        self.lazy = lazy
        self._connection = self._create_connection()

    @database_error_handler
    def execute(self, query: str, parameters: Dict[str, Any] = {}) -> None:
        """Executes Cypher query without returning any results."""
        cursor = self._connection.cursor()
        cursor.execute(query, parameters)
        cursor.fetchall()

    @database_error_handler
    def execute_and_fetch(self, query: str, parameters: Dict[str, Any] = {}) -> Iterator[Dict[str, Any]]:
        """Executes Cypher query and returns iterator of results."""
        cursor = self._connection.cursor()
        cursor.execute(query, parameters)
        while True:
            row = cursor.fetchone()
            if row is None:
                break
            yield {dsc.name: _convert_memgraph_value(row[index]) for index, dsc in enumerate(cursor.description)}

    def is_active(self) -> bool:
        """Returns True if connection is active and can be used."""
        return self._connection is not None and self._connection.status == mgclient.CONN_STATUS_READY

    @connection_handler
    def _create_connection(self) -> Connection:
        """Creates and returns a connection with Memgraph."""
        sslmode = mgclient.MG_SSLMODE_REQUIRE if self.encrypted else mgclient.MG_SSLMODE_DISABLE
        connection = mgclient.connect(
            host=self.host,
            port=self.port,
            username=self.username,
            password=self.password,
            sslmode=sslmode,
            lazy=self.lazy,
            client_name=self.client_name,
        )
        connection.autocommit = True
        return connection


def _convert_memgraph_value(value: Any) -> Any:
    """Converts Memgraph objects to custom Node/Relationship objects."""
    if isinstance(value, mgclient.Relationship):
        return Relationship.parse_obj(
            {
                "_type": value.type,
                "_id": value.id,
                "_start_node_id": value.start_id,
                "_end_node_id": value.end_id,
                **value.properties,
            }
        )

    if isinstance(value, mgclient.Node):
        return Node.parse_obj(
            {
                "_id": value.id,
                "_labels": set(value.labels),
                **value.properties,
            }
        )

    if isinstance(value, mgclient.Path):
        return Path.parse_obj(
            {
                "_nodes": list([_convert_memgraph_value(node) for node in value.nodes]),
                "_relationships": list([_convert_memgraph_value(rel) for rel in value.relationships]),
            }
        )

    return value


class Neo4jConnection(Connection):
    def __init__(
        self,
        host: str,
        port: int,
        username: str,
        password: str,
        encrypted: bool,
        client_name: Optional[str] = None,
        lazy: bool = True,
    ):
        super().__init__(
            host=host, port=port, username=username, password=password, encrypted=encrypted, client_name=client_name
        )
        self.lazy = lazy
        self._connection = self._create_connection()

    def execute(self, query: str, parameters: Dict[str, Any] = {}) -> None:
        """Executes Cypher query without returning any results."""
        with self._connection.session() as session:
            session.run(query, parameters)

    def execute_and_fetch(self, query: str, parameters: Dict[str, Any] = {}) -> Iterator[Dict[str, Any]]:
        """Executes Cypher query and returns iterator of results."""
        with self._connection.session() as session:
            results = session.run(query, parameters)
            columns = results.keys()
            for result in results:
                yield {column: _convert_neo4j_value(result[column]) for column in columns}

    def is_active(self) -> bool:
        """Returns True if connection is active and can be used."""
        return self._connection is not None

    def _create_connection(self):
        return GraphDatabase.driver(
            f"bolt://{self.host}:{self.port}", auth=(self.username, self.password), encrypted=self.encrypted
        )


def _convert_neo4j_value(value: Any) -> Any:
    """Converts Neo4j objects to custom Node/Relationship objects."""
    if isinstance(value, Neo4jRelationship):
        return Relationship.parse_obj(
            {
                "_type": value.type,
                "_id": value.id,
                "_start_node_id": value.start_node.id,
                "_end_node_id": value.end_node.id,
                **dict(value.items()),
            }
        )

    if isinstance(value, Neo4jNode):
        return Node.parse_obj(
            {
                "_id": value.id,
                "_labels": set(value.labels),
                **dict(value.items()),
            }
        )

    if isinstance(value, Neo4jPath):
        return Path.parse_obj(
            {
                "_nodes": list([_convert_neo4j_value(node) for node in value.nodes]),
                "_relationships": list([_convert_neo4j_value(rel) for rel in value.relationships]),
            }
        )

    return value
